from abc import ABC, abstractmethod
from typing import List


class MailerInterface(ABC):
    @abstractmethod
    def send_mail(self):
        ...


class Mailer(MailerInterface):
    def __init__(self, from_email):
        self.from_email = from_email

    def send_mail(self):
        print(f"Sending from {self.from_email}...")
        print("Sent")


class MultiMailer(MailerInterface):
    def __init__(self, mailers: List[MailerInterface]):
        self.mailers = mailers

    def send_mail(self):
        for mailer in self.mailers:
            mailer.send_mail()


class ProductService:
    def __init__(self, mailer: MailerInterface):
        self.mailer = mailer

    def process_product(self):
        print("processing product")
        self.mailer.send_mail()



from abc import ABC, abstractmethod

from oop_di import ContainerDefinition


# ### Domain code
class MailerInterface(ABC):
    @abstractmethod
    def send_mail(self):
        ...


class Mailer(MailerInterface):
    def __init__(self, from_email):
        self.from_email = from_email

    def send_mail(self):
        print(f"Sending from {self.from_email}...")
        print("Sent")


class ProductService:
    def __init__(self, mailer: MailerInterface):
        self.mailer = mailer

    def process_product(self):
        print("processing product")
        self.mailer.send_mail()


# ### Container definition
container_definition = ContainerDefinition()
container_definition.add_param("from_email", "test@example.com")
container_definition.add_service(ProductService)
container_definition.add_named_service(MailerInterface, Mailer)

container = container_definition.compile()


# ### Application code


@container.inject()
def process_product_endpoint_or_something(something, *, product_service: ProductService):
    print(something)
    product_service.process_product()


process_product_endpoint_or_something("doing something before calling product service")



from typing import Callable


class ServiceBuilder:
    def __init__(self, factory: Callable, *, is_singleton: bool):
        self.factory = factory
        self.is_singleton = is_singleton
        self.instance = None

    def get_instance(self):
        if not self.is_singleton:
            return self.factory()
        if self.instance:
            return self.instance
        self.instance = self.factory()

        return self.instance



import inspect
from functools import partial
from typing import Any, Callable, Dict, List

from .service_builder import ServiceBuilder
from .types import NameType


class Container:
    def __init__(self, params):
        self._services: Dict[NameType, ServiceBuilder] = {}
        self._params: Dict[NameType, Any] = params
        self._aliases: Dict[NameType, NameType] = {}
        self._by_tag: Dict[str, List[NameType]] = {}

    def add_alias(self, alias: NameType, service: NameType):
        self._aliases[alias] = service

    def has(self, name: NameType):
        if self.get(name, raise_if_none=False):
            return True

        return False

    def get(self, name: NameType, raise_if_none: bool = True):
        if isinstance(name, str) and name.startswith("#"):
            return self.get_tagged(name[1:])
        if alias := self._aliases.get(name):
            return self.get(alias)

        if service := self._services.get(name):
            return service.get_instance()

        if param := self._params.get(name):
            return param

        if raise_if_none:
            raise Exception(f"Cannot find {name}")

        return None

    def add_service(self, name: NameType, factory: Callable[[], Any], is_singleton: bool = True, tags: List[str] = []):
        self._services[name] = ServiceBuilder(factory, is_singleton=is_singleton)
        for tag in tags:
            try:
                self._by_tag[tag].append(name)
            except KeyError:
                self._by_tag[tag] = [name]

    def get_tagged(self, tag: str):
        names = self._by_tag.get(tag)
        if not names:
            return []

        return [self.get(name) for name in names]

    def inject(self, *, ignore_missing=True, **bindings):
        container = self

        def wrapper(f):
            args = inspect.getfullargspec(f)
            kwargs = {}
            for arg_name in args.kwonlyargs:
                if binding := bindings.get(arg_name):
                    value = container.get(binding, raise_if_none=not ignore_missing)
                    if value:
                        kwargs[arg_name] = value
                    continue

                try:
                    value = container.get(args.annotations[arg_name], raise_if_none=False) or container.get(
                        arg_name, raise_if_none=not ignore_missing
                    )
                    if value:
                        kwargs[arg_name] = value
                except KeyError:
                    value = container.get(arg_name, raise_if_none=not ignore_missing)
                    kwargs[arg_name] = value

            return partial(f, **kwargs)

        return wrapper



import pytest
from oop_di import ContainerDefinition, Extension
from oop_di.exception import CircularImportException, DefinitionNotFound


class SimpleTestService:
    ...


class DimpleTestService(SimpleTestService):
    ...


class DependentService:
    def __init__(self, simple: SimpleTestService):
        self.simple = simple


class CircularImportProblem1:
    def __init__(self, mention_it_self: "CircularImportProblem1"):
        self.mention_it_self = mention_it_self


def simple_factory(simple: SimpleTestService):
    return isinstance(simple, SimpleTestService)


def simple_factory_for_tagged(tagged):
    return len(tagged)


class TestContainerDefinition:
    def setup_method(self):
        self.sut = ContainerDefinition()

    def test_it_should_compile_simple_service(self):
        self.sut.add_service(SimpleTestService)
        container = self.sut.compile()
        service = container.get(SimpleTestService)
        assert isinstance(service, SimpleTestService)

    def test_it_should_compile_named_service(self):
        self.sut.add_named_service("xxx", SimpleTestService)
        container = self.sut.compile()
        service = container.get("xxx")
        assert isinstance(service, SimpleTestService)

    def test_it_should_compile_dependent_service(self):
        self.sut.add_service(SimpleTestService)
        self.sut.add_service(DependentService)
        container = self.sut.compile()
        service = container.get(DependentService)
        assert isinstance(service, DependentService)

    def test_it_should_detect_circular_import_problem(self):
        self.sut.add_named_service("CircularImportProblem1", CircularImportProblem1)
        with pytest.raises(CircularImportException):
            self.sut.compile()

    def test_it_should_notify_if_dependency_not_registered(self):
        self.sut.add_service(DependentService)
        with pytest.raises(DefinitionNotFound):
            self.sut.compile()

    def test_it_should_compile_factory(self):
        self.sut.add_service(SimpleTestService)
        self.sut.add_factory("test", simple_factory)
        container = self.sut.compile()
        assert container.get("test")

    def test_it_should_compile_param(self):
        self.sut.add_param("test", "y")
        container = self.sut.compile()
        assert "y" == container.get("test")

    def test_it_should_inject_by_name_if_by_type_fails(self):
        self.sut.add_service(DependentService)
        self.sut.add_named_service("simple", SimpleTestService)
        container = self.sut.compile()
        assert isinstance(container.get(DependentService), DependentService)

    def test_it_should_inject_binding_by_name(self):
        self.sut.add_service(DependentService, simple="dimple")
        self.sut.add_service(SimpleTestService)
        self.sut.add_named_service("dimple", DimpleTestService)
        container = self.sut.compile()
        assert isinstance(container.get(DependentService).simple, DimpleTestService)

    def test_it_should_compile_aliases(self):
        self.sut.add_service(DependentService, simple="dimple")
        self.sut.add_service(SimpleTestService)
        self.sut.add_service(DimpleTestService)
        self.sut.add_alias("dimple", DimpleTestService)
        container = self.sut.compile()
        assert isinstance(container.get(DependentService).simple, DimpleTestService)

    def test_it_should_pass_singleton_flag(self):
        self.sut.add_service(SimpleTestService)
        self.sut.add_named_service("x", SimpleTestService, is_singleton=False)
        container = self.sut.compile()
        assert container.get(SimpleTestService) == container.get(SimpleTestService)
        assert container.get("x") != container.get("x")
        assert isinstance(container.get("x"), SimpleTestService)

    def test_it_should_inject_tagged_services(self):
        self.sut.add_named_service("x", SimpleTestService, tags=["tag1", "tag2"])
        self.sut.add_named_service("y", SimpleTestService, tags=["tag1"])
        self.sut.add_named_service("z", SimpleTestService, tags=["tag1", "tag2"])
        self.sut.add_factory("threetags", simple_factory_for_tagged, tagged="#tag1")
        self.sut.add_factory("twotags", simple_factory_for_tagged, tagged="#tag2")
        container = self.sut.compile()
        assert 3 == container.get("threetags")
        assert 2 == container.get("twotags")

    def test_it_should_combine_extensions(self):
        class Ext1(Extension):
            def define(self):
                self.add_service(SimpleTestService, tags=["tag1"])

        class Ext2(Extension):
            def define(self):
                self.add_service(DependentService, tags=["tag1"])

        self.sut.add_factory("twotags", simple_factory_for_tagged, tagged="#tag1")
        self.sut.add_extension(Ext1())
        self.sut.add_extension(Ext2())

        container = self.sut.compile()
        assert isinstance(container.get(DependentService), DependentService)
        assert 2 == container.get("twotags")



from .container import Container
from .container_definition import ContainerDefinition
from .extension import EnvExtension, Extension, JsonExtension

__all__ = ["Container", "ContainerDefinition", "Extension", "EnvExtension", "JsonExtension"]



from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List

from ..definition import Definition
from ..types import NameType


class Extension(ABC):
    def __init__(self):
        self._definitions: Dict[NameType, Definition] = {}
        self._params: Dict[NameType, Any] = {}
        self._aliases: Dict[NameType, NameType] = {}
        self.define()

    def add_alias(self, alias: NameType, service: NameType):
        self._aliases[alias] = service

    def add_service(self, service: type, *, tags: List[str] = [], is_singleton=True, **bind):
        self._definitions[service] = Definition(service, service, bind, is_singleton=is_singleton, tags=tags)

    def add_named_service(self, name: NameType, service: type, *, tags: List[str] = [], is_singleton=True, **bind):
        self._definitions[name] = Definition(name, service, bind, is_singleton=is_singleton, tags=tags)

    def add_factory(self, name: NameType, factory: Callable, *, tags: List[str] = [], is_singleton=True, **bind):
        self._definitions[name] = Definition(name, factory, bind, is_singleton=is_singleton, tags=tags)

    def add_param(self, name: NameType, value: Any):
        self._params[name] = value

    def get_definitions(self) -> Dict[NameType, Definition]:
        return self._definitions

    def get_aliases(self) -> Dict[NameType, NameType]:
        return self._aliases

    def get_params(self) -> Dict[NameType, Any]:
        return self._params

    @abstractmethod
    def define(self):
        ...



from pathlib import Path

from oop_di import ContainerDefinition, JsonExtension

from examples.s4_module import ProductService

container_definition = ContainerDefinition()
container_definition.add_extension(JsonExtension(Path(__file__).parent / "s4_config.json"))
container = container_definition.compile()


@container.inject()
def process_product_endpoint(something, *, product_service: ProductService):
    print(something)
    product_service.process_product()


process_product_endpoint("doing something before calling product service")



import os
from unittest import mock

from oop_di.extension import EnvExtension


def mockenv(**envvars):
    return mock.patch.dict(os.environ, envvars)


class TestEnvExtension:
    @mockenv(OLOLO="TROLOLO", XXX="yyy")
    def test_it_should_add_env_as_params(self):
        sut = EnvExtension()
        params = sut.get_params()
        assert params["ololo"] == "TROLOLO"
        assert params["xxx"] == "yyy"

    @mockenv(OLOLO="TROLOLO", xxx="yyy")
    def test_it_should_add_env_as_params_without_changing_case(self):
        sut = EnvExtension(keys_to_lower=False)
        params = sut.get_params()
        assert params["OLOLO"] == "TROLOLO"
        assert params["xxx"] == "yyy"



from pathlib import Path

from oop_di.extension import JsonExtension
from tests.extension.json.services import Service1, Service2, Service3


class TestJsonExtension:
    def test_it_should_load_config(self):
        extension = JsonExtension(Path(__file__).parent / "config.json")

        params = extension.get_params()
        aliases = extension.get_aliases()
        services = extension.get_definitions()

        assert params == {"ololo": "trololo", "xxx": "yyy"}
        assert aliases == {"zzz": "xxx", "s2": Service2}
        service1 = services["s1"]
        service2 = services[Service2]
        service3 = services[Service3]

        assert service1.tags == ["a"]
        assert service2.tags == ["a", "b"]
        assert service3.tags == []

        assert service1.bindings == {"parameter1": "xxx"}
        assert service2.bindings == {}
        assert service3.bindings == {"parameter": Service1}



from oop_di import Container


class TestContainer:
    def setup_method(self):
        self.sut = Container({"x": "y"})

    def test_it_should_get_param(self):
        assert "y" == self.sut.get("x")

    def test_it_should_get_service(self):
        self.sut.add_service(Container, lambda: "lol")
        self.sut.add_service("test", lambda: "lol2")
        assert "lol" == self.sut.get(Container)
        assert "lol2" == self.sut.get("test")

    def test_it_should_resolve_alias(self):
        self.sut.add_service(Container, lambda: "lol")
        self.sut.add_alias("test", Container)
        assert "lol" == self.sut.get("test")

    def test_it_should_get_by_tags(self):
        self.sut.add_service("a", lambda: "lolA", tags=["tag1", "tag2"])
        self.sut.add_service("b", lambda: "lolB", tags=["tag1"])
        self.sut.add_service("c", lambda: "lolC", tags=["tag2"])
        assert ["lolA", "lolB"] == self.sut.get_tagged("tag1")
        assert ["lolA", "lolC"] == self.sut.get_tagged("tag2")

    def test_it_should_inject_kwargs(self):
        self.sut.add_service(Container, lambda: "lol")

        @self.sut.inject(t="x")
        def a(s, *, y: Container, t: int, define_it):
            return [s, y, t, define_it]

        assert ["test", "lol", "y", "defined"] == a("test", define_it="defined")



from typing import Callable, Dict, List

from .types import NameType


class Definition:
    def __init__(
        self, name: NameType, factory: Callable, bindings: Dict[str, NameType], is_singleton: bool, tags: List[str]
    ):
        self.name = name
        self.factory = factory
        self.bindings = bindings
        self.is_singleton = is_singleton
        self.tags = tags



import inspect
from typing import Any, Callable, Dict, List

from .container import Container
from .definition import Definition
from .exception import CircularImportException, DefinitionNotFound
from .extension import Extension
from .types import NameType


class ContainerDefinition:
    def __init__(self):
        self._definitions: Dict[NameType, Definition] = {}
        self._params: Dict[NameType, Any] = {}
        self._in_compile: Dict[NameType, bool] = {}
        self._aliases: Dict[NameType, NameType] = {}

    def add_alias(self, alias: NameType, service: NameType):
        self._aliases[alias] = service

    def add_service(self, service: type, *, tags: List[str] = [], is_singleton=True, **bind):
        self._definitions[service] = Definition(service, service, bind, is_singleton=is_singleton, tags=tags)

    def add_named_service(self, name: NameType, service: type, *, tags: List[str] = [], is_singleton=True, **bind):
        self._definitions[name] = Definition(name, service, bind, is_singleton=is_singleton, tags=tags)

    def add_factory(self, name: NameType, factory: Callable, *, tags: List[str] = [], is_singleton=True, **bind):
        self._definitions[name] = Definition(name, factory, bind, is_singleton=is_singleton, tags=tags)

    def add_param(self, name: NameType, value: Any):
        self._params[name] = value

    def add_extension(self, extension: Extension):
        self._definitions.update(extension.get_definitions())
        self._aliases.update(extension.get_aliases())
        self._params.update(extension.get_params())

    def compile(self) -> Container:
        container = Container(self._params)
        for name in self._definitions.keys():
            self._compile_definition(name, container)

        for alias, service in self._aliases.items():
            container.add_alias(alias, service)

        return container

    def _compile_definition(self, name, container: Container):
        if name in self._aliases:
            return self._compile_definition(self._aliases[name], container)
        if name in self._params:
            return

        definition = self._definitions.get(name)
        if not definition:
            raise DefinitionNotFound(f"No definition for {name} found")
        if self._in_compile.get(definition.name, None):
            raise CircularImportException("Circular import error")

        self._in_compile[definition.name] = True
        if container.has(definition.name):
            self._in_compile.pop(definition.name)
            return

        self._resolve_dependencies(definition, container)
        container.add_service(
            definition.name,
            self._bind_factory_to_container(definition, container),
            definition.is_singleton,
            definition.tags,
        )

        self._in_compile.pop(definition.name)

    def _bind_factory_to_container(self, definition: Definition, container: Container):
        factory = definition.factory
        bindings = definition.bindings
        args = inspect.getfullargspec(factory)

        def bound_factory():
            arguments = []
            for arg_name in args.args:
                if arg_name == "self":
                    continue
                if binding := bindings.get(arg_name):
                    arguments.append(container.get(binding))
                    continue
                try:
                    arguments.append(
                        container.get(args.annotations[arg_name], raise_if_none=False) or container.get(arg_name)
                    )
                except KeyError:
                    arguments.append(container.get(arg_name))
            kwargs = {}
            for arg_name in args.kwonlyargs:
                if binding := bindings.get(arg_name):
                    kwargs[arg_name] = container.get(binding)
                    continue
                if args.kwonlydefaults and args.kwonlydefaults.get(arg_name):
                    d = args.kwonlydefaults.get(arg_name)
                    kwargs[arg_name] = d
                    continue

                try:
                    kwargs[arg_name] = container.get(args.annotations[arg_name], raise_if_none=False) or container.get(
                        arg_name
                    )
                except KeyError:
                    kwargs[arg_name] = container.get(arg_name)

            return factory(*arguments, **kwargs)

        return bound_factory

    def _resolve_dependencies(self, definition: Definition, container: Container):
        factory = definition.factory
        bindings = definition.bindings
        args = inspect.getfullargspec(factory)

        for arg_name in args.args:
            if arg_name == "self":
                continue

            if binding := bindings.get(arg_name):
                if isinstance(binding, str) and binding.startswith("#"):
                    continue
                self._compile_definition(binding, container)
                continue

            try:
                self._compile_definition(args.annotations[arg_name], container)
            except (KeyError, DefinitionNotFound):
                self._compile_definition(arg_name, container)
        for arg_name in args.kwonlyargs:
            if binding := bindings.get(arg_name):
                if isinstance(binding, str) and binding.startswith("#"):
                    continue
                self._compile_definition(binding, container)
                continue
            if args.kwonlydefaults and args.kwonlydefaults.get(arg_name):
                continue

            try:
                self._compile_definition(args.annotations[arg_name], container)
            except (KeyError, DefinitionNotFound):
                self._compile_definition(arg_name, container)



from abc import ABC, abstractmethod
from typing import List

from oop_di import ContainerDefinition


class MailerInterface(ABC):
    @abstractmethod
    def send_mail(self):
        ...


class Mailer(MailerInterface):
    def __init__(self, from_email):
        self.from_email = from_email

    def send_mail(self):
        print(f"Sending from {self.from_email}...")
        print("Sent")


class MultiMailer(MailerInterface):
    def __init__(self, mailers: List[MailerInterface]):
        self.mailers = mailers

    def send_mail(self):
        for mailer in self.mailers:
            mailer.send_mail()


class ProductService:
    def __init__(self, mailer: MailerInterface):
        self.mailer = mailer

    def process_product(self):
        print("processing product")
        self.mailer.send_mail()


container_definition = ContainerDefinition()
container_definition.add_service(ProductService)

container_definition.add_param("test1_email", "test@example.com")
container_definition.add_named_service("admin1_mailer", Mailer, from_email="test1_email", tags=["admin_mailers"])
container_definition.add_param("test2_email", "test2@example.com")
container_definition.add_named_service("admin2_mailer", Mailer, from_email="test2_email", tags=["admin_mailers"])
container_definition.add_param("test3_email", "test3@example.com")
container_definition.add_named_service("admin3_mailer", Mailer, from_email="test3_email", tags=["admin_mailers"])

container_definition.add_named_service(MailerInterface, MultiMailer, mailers="#admin_mailers")

container = container_definition.compile()


@container.inject()
def process_product_endpoint(something, *, product_service: ProductService):
    print(something)
    product_service.process_product()


process_product_endpoint("doing something before calling product service")



import codecs
import os

from setuptools import find_packages, setup


def open_local(path, mode="r", encoding="utf8"):
    path = os.path.join(os.path.abspath(os.path.dirname(__file__)), path)
    return codecs.open(path, mode, encoding)


long_description = ""
with open_local("README.md") as fp:
    long_description += fp.read()
    long_description += "\n"


setup(
    name="oop-di",
    url="https://github.com/ChubV/oop-di/",
    license="MIT",
    description="A simple OOP Dependency injection container",
    author="Vladimir Chub",
    author_email="v@chub.com.ua",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="utils",
    python_requires=">=3.9",
    packages=find_packages(exclude=("tests", "examples")),
    install_requires=None,
    include_package_data=True,
    tests_require=["pytest", "pytest-asyncio"],
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
)



import os

from .extension import Extension


class EnvExtension(Extension):
    def __init__(self, keys_to_lower: bool = True):
        self.keys_to_lower = keys_to_lower
        super().__init__()

    def define(self):
        for k, v in os.environ.items():
            self.add_param(k.lower() if self.keys_to_lower else k, v)



from oop_di.service_builder import ServiceBuilder


class TestServiceBuilder:
    def test_it_should_use_factory_to_get_instance(self):
        service = ServiceBuilder(lambda: "ololo", is_singleton=True)
        assert "ololo" == service.get_instance()

    def test_it_should_cache_the_instance(self):
        service = ServiceBuilder(lambda: TestServiceBuilder(), is_singleton=True)
        assert service.get_instance() == service.get_instance()

    def test_it_should_not_cache_the_instance(self):
        service = ServiceBuilder(lambda: TestServiceBuilder(), is_singleton=False)
        assert service.get_instance() != service.get_instance()



import importlib
from json import load
from pathlib import Path
from typing import Any, Dict, List

from ..types import NameType
from .extension import Extension


class JsonExtension(Extension):
    def __init__(self, json_path: Path):
        self.json_path = json_path
        super().__init__()

    def define(self):
        with open(self.json_path) as file:
            data = load(file)

        self._load_parameters(data.get("parameters", {}))
        self._load_aliases(data.get("aliases", {}))
        self._load_services(data.get("services", {}))
        importlib.invalidate_caches()

    def _load_parameters(self, data: Dict[str, Any]):
        for k, v in data.items():
            self.add_param(self._resolve_key(k), v)

    def _load_aliases(self, data: Dict[str, Any]):
        for k, v in data.items():
            self.add_alias(self._resolve_key(k), self._resolve_key(v))

    def _load_services(self, data: Dict[str, Any]):
        for k, v in data.items():
            self._load_module(k, v)

    def _load_module(self, mod: str, data: List[Dict[str, Any]]):
        for service in data:
            self._load_service(mod, service["class"], service)

    def _load_service(self, mod: str, item: str, data: Dict[str, Any]):
        service = self._resolve_key(f"@{mod}.{item}")
        parameters = {}
        for k, v in data.get("parameters", {}).items():
            parameters[k] = self._resolve_key(v)

        if name := data.get("name"):
            self.add_named_service(
                self._resolve_key(name),
                service,
                tags=data.get("tags", []),
                is_singleton=data.get("is_singleton", True),
                **parameters,
            )
        else:
            self.add_service(
                service, tags=data.get("tags", []), is_singleton=data.get("is_singleton", True), **parameters
            )
        for alias in data.get("aliases", []):
            self.add_alias(alias, service)

    def _resolve_key(self, key: str) -> NameType:
        if not key.startswith("@"):
            return key

        mod, class_name = key[1:].rsplit(".", 1)
        loaded = importlib.import_module(mod)

        return getattr(loaded, class_name)
